// n505ve_audio 1999-12-20
/* Copyright (c) 1999 Andrew Green. ag@electricmagic.com
Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files  (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/
// http://www.electricmagic.com/~ag/snippets/n505ve_audio.html
// Portions derived from Joseph Wang's sbpro and Be Inc.'s sonic_vibes sample driver.
#include "n505ve_audio.h"

#include "audio_driver.h" // for audio_format
#include <string.h> // for memset
#include <algobase.h> // for min


//#define MAX_PORT_VALUE 0x350
#define DEFAULT_PORT 0x220
#define DEFAULT_IRQ 5
#define DEFAULT_DMA8 1
//#define DEFAULT_DMA16 5
//#define DEFAULT_MPU 0x320
#define SBPRO_DELAY 3

// Annotations of constants pulled from Yamaha's specs for YMF744B, Section 2.2.
// Also checked against values in Linux Sound Blaster driver.
// Port offsets
#define SBPRO_REG 0x4 // Mixer address port
#define SBPRO_DATA 0x5 // Mixer data port
#define SBPRO_RESET 0x6 // Reset port
#define SBPRO_READ_DATA 0xA // Read data port
#define SBPRO_WRITE_DATA 0xC // Write command/data port (also write buffer status port)
#define SBPRO_WRITE_BUFFER_STATUS 0xC // Write buffer status port (also write command/data port)
#define SBPRO_READ_BUFFER_STATUS 0xE // Read buffer status port

// Commands
#define SBPRO_TIME_CONSTANT 0x40 // Set digitized sound transfer Time Constant
#define SBPRO_TRANSFER_SIZE 0x48 // Set DSP block transfer size
#define SBPRO_PLAYBACK_8_BIT 0x90 // 8 bit high-speed auto-init DMA mode digitzed sound output
#define SBPRO_EXIT_DMA 0xD0 // Pause 8 bit DMA mode digitized sound I/O
#define SBPRO_SPEAKER_ENABLE 0xD1 // Turn on speaker
#define SBPRO_SPEAKER_DISABLE 0xD3 // Turn off speaker
#define SBPRO_VERSION 0xE1 // Get DSP version number

// Mixer commands
#define SBPRO_VOICE 0x04
#define SBPRO_MIC 0x0A
#define SBPRO_INPUT_MUX 0x0C
#define SBPRO_OUTPUT_MUX 0x0E
#define SBPRO_MASTER 0x22
#define SBPRO_MIDI 0x26
#define SBPRO_CD 0x28
#define SBPRO_LINE 0x2E

// ==================================================
// Our implementations of sInitHardware and sInitDriver, needed to bootstrap the object-based model.
status_t ZDriver_Be_Driver::sInitHardware()
	{
	return B_OK;
	}

status_t ZDriver_Be_Driver::sInitDriver()
	{
	N505VE_Audio_Driver* theDriver = new N505VE_Audio_Driver;
	if (theDriver->Initialize())
		return B_OK;
	theDriver->Finalize();
	delete theDriver;
	return B_ERROR;
	}

// ==================================================

N505VE_Audio_Driver::N505VE_Audio_Driver()
	{}

N505VE_Audio_Driver::~N505VE_Audio_Driver()
	{}

bool N505VE_Audio_Driver::Initialize()
	{
	if (!ZDriver_Be_Driver::Initialize())
		return false;
	if (!this->AddCard(new N505VE_Audio_Card(DEFAULT_PORT, DEFAULT_IRQ, DEFAULT_DMA8)))
		return false;
	return true;
	}

void N505VE_Audio_Driver::Finalize()
	{
	ZDriver_Be_Driver::Finalize();
	}

// ====================================================================================================

N505VE_Audio_Card::N505VE_Audio_Card(int inPort, int inIRQ, int inDMA8)
	{
	fPort = inPort;
	fIRQ = inIRQ;
	fDMA8 = inDMA8;
	}

N505VE_Audio_Card::~N505VE_Audio_Card()
	{
	}

bool N505VE_Audio_Card::Initialize()
	{
	if (!ZDriver_Be_Card::Initialize())
		return false;

	if (!this->AddDevice(new N505VE_Audio_DevicePCM(fPort, fIRQ, fDMA8)))
		return false;

	return true;
	}

void N505VE_Audio_Card::Finalize()
	{
	ZDriver_Be_Card::Finalize();
	}


// ====================================================================================================

N505VE_Audio_DevicePCM::N505VE_Audio_DevicePCM(int inPort, int inIRQ, int inDMA8)
	{
	fCountInterruptHandler = 0;
	fSpinLock = 0;

	fWriteBuffer = nil;
	fWriteBufferSize = nil;
	fWriteBufferArea = nil;

	fDoubleBufferSize = 0;
	fDoubleBuffer[0] = nil;
	fDoubleBuffer[1]= nil;

	fSem_PlaybackComplete = -1;
	fSem_Write = -1;
	fSem_Init = -1;

	fCount_WriteWaiting = 0;

	fCount_Open = 0;

	fSampleClock = 0;
	fSampleTime = 0;

	fISAModuleInfo = nil;

	fPort = inPort;
	fIRQ = inIRQ;
	fDMA8 = inDMA8;
	}

N505VE_Audio_DevicePCM::~N505VE_Audio_DevicePCM()
	{
	}

bool N505VE_Audio_DevicePCM::Initialize()
	{
// Now grab the resources we'll need
	if (::get_module(B_ISA_MODULE_NAME, (module_info**)&fISAModuleInfo) != B_OK)
		return false;

// Check that the hardware is okay
	bool resetWorked = false;
	cpu_status priorCPUStatus = this->DisableInterruptsAndAcquireSpinLock();
	for (int i = 33; i > 0; --i)
		{
// poll dsp for 100usec, 3usec each poll
		this->WriteIO(SBPRO_RESET, 1);
		::spin(SBPRO_DELAY);
		this->WriteIO(SBPRO_RESET, 0);

		uint8 value;
		if (this->ReadData(value) == B_OK)
			{
			if (value == 0xAA)
				{
				resetWorked = true;
				break;
				}
			}
		}
	this->ReleaseSpinLockAndRestoreInterrupts(priorCPUStatus);

	if (!resetWorked)
		return false;

// get the dsp version number
	if (this->WriteData(SBPRO_VERSION) != B_OK)
		return false;
	uint8 high;
	if (this->ReadData(high) != B_OK)
		return false;
	uint8 low;
	if (this->ReadData(low) != B_OK)
		return false;
	if (high < 3)
		return false;

	fISAModuleInfo->lock_isa_dma_channel(fDMA8);
	fLockedDMA = true;

	fSem_Init = ::create_sem(1, "N505VE_Audio_DevicePCM::fSem_Init");
	if (!fSem_Init < B_OK)
		return false;

	fWriteBufferSize = 1024;

	fWriteBufferArea = ::create_area("n505ve_audio playback buffer", &fWriteBuffer, B_ANY_KERNEL_ADDRESS, B_PAGE_SIZE, B_LOMEM, B_READ_AREA | B_WRITE_AREA);
	if (fWriteBufferArea == nil)
		return false;

	physical_entry table;
	if (::get_memory_map(fWriteBuffer, fWriteBufferSize, &table, 1) < B_OK)
		return false;

	if (((uint32)table.address) & 0xff000000)
		return false;
	if ((((uint32)table.address) + fWriteBufferSize) & 0xff000000)
		return false;

	fDoubleBufferSize = fWriteBufferSize/2;
	fDoubleBuffer[0] = reinterpret_cast<unsigned char*>(fWriteBuffer);
	fDoubleBuffer[1] = reinterpret_cast<unsigned char*>(fWriteBuffer) + fDoubleBufferSize;

	fDoubleBufferCurrent = 0;

	::memset(fDoubleBuffer[0], 0x80, fDoubleBufferSize);
	::memset(fDoubleBuffer[1], 0x80, fDoubleBufferSize);

	return true;
	}

void N505VE_Audio_DevicePCM::Finalize()
	{
	if (fWriteBufferArea)
		{
		fDoubleBufferSize = 0;
		fDoubleBuffer[0] = nil;
		fDoubleBuffer[1] = nil;
		::delete_area(fWriteBufferArea);
		fWriteBufferArea = nil;
		}

	if (fSem_Init >= 0)
		{
		::delete_sem(fSem_Init);
		fSem_Init = -1;
		}

	if (fLockedDMA)
		{
	   	fISAModuleInfo->unlock_isa_dma_channel(fDMA8);
		fLockedDMA = false;
		}

	if (fISAModuleInfo)
		{
		::put_module(B_ISA_MODULE_NAME);
		fISAModuleInfo = nil;
		}
	}

size_t N505VE_Audio_DevicePCM::CountNames()
	{
	return 2;
	}

static const char* sDeviceNames[] =
	{
	"audio/raw/n505ve_audio/1",
	"audio/old/n505ve_audio/1",
	nil
	};

const char* N505VE_Audio_DevicePCM::GetNameByIndex(size_t inNameIndex)
	{
//	ZAssert(inNameIndex < 2);
	return sDeviceNames[inNameIndex];
	}

status_t N505VE_Audio_DevicePCM::Open(const char* inName, uint32 inFlags, ZDriver_Be_DeviceInstance*& outDeviceInstance)
	{
	::acquire_sem(fSem_Init);
	
	if (::atomic_add(&fCount_Open, 1) == 0)
		{
		fSem_Write = ::create_sem(1, "N505VE_Audio_DevicePCM::fSem_Write");
		if (fSem_Write < B_OK)
			goto bailout;
		::set_sem_owner(fSem_Write, B_SYSTEM_TEAM);

		fSem_WriteEntry = ::create_sem(1, "N505VE_Audio_DevicePCM::fSem_WriteEntry");
		if (fSem_WriteEntry < B_OK)
			goto bailout;
		::set_sem_owner(fSem_WriteEntry, B_SYSTEM_TEAM);

		this->IncrementInterruptHandler();
		this->StartDMA();
		}
	else
		{
		}
	outDeviceInstance = this;
	::release_sem(fSem_Init);
	return B_OK;

bailout:
	if (fSem_WriteEntry >= 0)
		{
		::delete_sem(fSem_WriteEntry);
		fSem_WriteEntry = -1;
		}
	if (fSem_Write >= 0)
		{
		::delete_sem(fSem_Write);
		fSem_Write = -1;
		}

	::atomic_add(&fCount_Open, -1);
	::release_sem(fSem_Init);
	return B_ERROR;
	}

status_t N505VE_Audio_DevicePCM::Close()
	{
	::acquire_sem(fSem_Init);
	if (::atomic_add(&fCount_Open, -1) == 1)
		{
		this->StopDMA();
		this->DecrementInterruptHandler();		

		::delete_sem(fSem_WriteEntry);
		fSem_WriteEntry = -1;
		::delete_sem(fSem_Write);
		fSem_Write = -1;
		}

	::release_sem(fSem_Init);
	return B_OK;
	}

status_t N505VE_Audio_DevicePCM::Free()
	{
	::acquire_sem(fSem_Init);
	if (fCount_Open == 0)
		{
// Take any last actions
		}
	::release_sem(fSem_Init);
	return B_OK;
	}

status_t N505VE_Audio_DevicePCM::Control(uint32 inOperation, void* inOutData, size_t inLen)
	{
	static float rates[] = { 44100.0, 44100.0, 44100.0, 44100.0 };
	switch (inOperation)
		{
		case B_AUDIO_GET_AUDIO_FORMAT:
			{
			audio_format* theFormat = reinterpret_cast<audio_format*>(inOutData);
			theFormat->sample_rate = 44100;
			theFormat->channels = 2;
			theFormat->format = 0x02; // 8 bit samples
			theFormat->big_endian = 0;
			theFormat->buf_header = 0;
			theFormat->play_buf_size = 2048;
			theFormat->rec_buf_size = 0;
			return B_OK;
			}
		case B_AUDIO_GET_PREFERRED_SAMPLE_RATES:
			{
			::memcpy(inOutData, rates, sizeof(rates));
			return B_OK;
			}
		case SOUND_GET_PARAMS:
			{
			return this->GetParams(*reinterpret_cast<sound_setup*>(inOutData));
			}
		case SOUND_SET_PARAMS:
			{
			return this->SetParams(*reinterpret_cast<sound_setup*>(inOutData));
			}
		case SOUND_SET_PLAYBACK_COMPLETION_SEM:
			{
			fSem_PlaybackComplete = *reinterpret_cast<sem_id*>(inOutData);
			return B_OK;
			}
			break;
		case SOUND_UNSAFE_WRITE: 
			{
			audio_buffer_header* header = reinterpret_cast<audio_buffer_header*>(inOutData);
			size_t realLen = header->reserved_1 - sizeof(audio_buffer_header);

			this->Internal_Write(&header[1], realLen, header->time, header->sample_clock);

			return release_sem(fSem_PlaybackComplete);
			}
			break;
		case SOUND_GET_PLAYBACK_PREFERRED_BUF_SIZE:
			{
			*reinterpret_cast<int32*>(inOutData) = 2048;
			return B_OK;
			}
		}
	return EINVAL;
	}

status_t N505VE_Audio_DevicePCM::Read(off_t inPos, void* outData, size_t& inOutLen)
	{
	inOutLen = 0;
	return B_ERROR;
	}

status_t N505VE_Audio_DevicePCM::Write(off_t inPos, const void* inData, size_t& inOutLen)
	{
	bigtime_t dummyTime, dummySampleTime;
	return this->Internal_Write(inData, inOutLen, dummyTime, dummySampleTime);
	}

status_t N505VE_Audio_DevicePCM::Internal_Write(const void* inData, size_t& inOutLen, bigtime_t& outTime, bigtime_t& outSampleTime)
	{
	if (::acquire_sem_etc(fSem_WriteEntry, 1, B_CAN_INTERRUPT, 0) < B_OK)
		return B_ERROR;

	short* localSource = reinterpret_cast<short*>(inData);
	size_t destSampleCount = inOutLen / sizeof(short);

	bool isFirst = true;
	while (destSampleCount > 0)
		{
// Wait till the next buffer flip
		::atomic_add(&fCount_WriteWaiting, 1);
		if (::acquire_sem_etc(fSem_Write, 1, B_CAN_INTERRUPT, 0) < B_OK)
			{
// Bail if our semaphore was deleted.
			::release_sem(fSem_WriteEntry);
			return B_ERROR;
			}

		if (isFirst)
			{
// Record the real time and sample clock of the buffer that was just
// handled by our interrupt routine, but only on the first time through this loop.
			outTime = fSampleTime;
			outSampleTime = fSampleClock / 2 * 10000 / 441;
			isFirst = false;
			}

// Identify which buffer to write to, and how much to move on this iteration.
		unsigned char* bufferToWrite = fDoubleBuffer[1 - fDoubleBufferCurrent];
		size_t samplesToMove = min(destSampleCount, fDoubleBufferSize);

// Write silence to the tail of the buffer, if we have insufficient data to fill it.
		if (samplesToMove < fDoubleBufferSize)
			memset(bufferToWrite + samplesToMove, 0x80, fDoubleBufferSize - samplesToMove);

// Move the samples. We add one to the count so we can control looping with a pre-decrement,
// which is faster on some architectures (is this a win on Intel?) We divide by two and do
// two identical copies to make it clear we're moving the left channel and the right channel
// independently.
		size_t count = samplesToMove / 2 + 1;
		while (--count)
			{
// Left channel
			*bufferToWrite++ = ((*localSource++) >> 8) ^ 0x80;
// Right channel
			*bufferToWrite++ = ((*localSource++) >> 8) ^ 0x80;
			}

		destSampleCount -= samplesToMove;
		}	

	::release_sem(fSem_WriteEntry);
	return B_OK;
	}


status_t N505VE_Audio_DevicePCM::SetParams(const sound_setup& inSoundSetup)
	{
	cpu_status priorStatus;
	int value;

// aux1 maps to CD-ROM, 5bit->4bit
	value = 0;
	if (!inSoundSetup.left.aux1_mix_mute)
		value += 0xE0 - ((inSoundSetup.left.aux1_mix_gain << 3) & 0xE0);
	if (!inSoundSetup.right.aux1_mix_mute)
		value += 0x0E - ((inSoundSetup.right.aux1_mix_gain >> 1) & 0x0E);
	priorStatus = this->DisableInterruptsAndAcquireSpinLock();
	this->WriteReg(SBPRO_CD, value);
	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);

// aux2 maps to MIDI, 5->4
	value = 0;
	if (!inSoundSetup.left.aux2_mix_mute)
		value += 0xE0 - ((inSoundSetup.left.aux2_mix_gain << 3) & 0xE0);
	if (!inSoundSetup.right.aux2_mix_mute)
		value += 0x0E - ((inSoundSetup.right.aux2_mix_gain >> 1) & 0x0E);
	priorStatus = this->DisableInterruptsAndAcquireSpinLock();
	this->WriteReg(SBPRO_MIDI, value);
	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);

// set line levels, 5->4
	value = 0;
	if (!inSoundSetup.left.line_mix_mute)
		value += 0xE0 - ((inSoundSetup.left.line_mix_gain << 3) & 0xE0);
	if (!inSoundSetup.right.line_mix_mute)
		value += 0x0E - ((inSoundSetup.right.line_mix_gain >> 1) & 0x0E);
	priorStatus = this->DisableInterruptsAndAcquireSpinLock();
	this->WriteReg(SBPRO_LINE, value);
	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);

// dac maps to voice, 6->4
	value = 0;
	if (!inSoundSetup.left.dac_mute)
		value += 0xE0 - ((inSoundSetup.left.dac_attn << 2) & 0xE0);
	if (!inSoundSetup.right.dac_mute)
		value += 0x0E - ((inSoundSetup.right.dac_attn >> 2) & 0x0E);
	priorStatus = this->DisableInterruptsAndAcquireSpinLock();
	this->WriteReg(SBPRO_VOICE, value);
	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);

	return B_OK;
	}

status_t N505VE_Audio_DevicePCM::GetParams(sound_setup& outSoundSetup)
	{
	int theValue;

// no controls for channel.adc_gain
	outSoundSetup.left.adc_gain = 15;
	outSoundSetup.right.adc_gain = 15;

// set channel.mic levels, 1bit->2bit
	outSoundSetup.left.mic_gain_enable = 0;
	outSoundSetup.right.mic_gain_enable = 0;

// DAC maps to voice, 6->4
	theValue = this->ReadReg(SBPRO_VOICE);
	outSoundSetup.left.dac_attn = 0x38 - ((theValue >> 2) & 0x38);
	outSoundSetup.right.dac_attn = 0x38 - ((theValue << 2) & 0x38);
	outSoundSetup.left.dac_mute = 0;
	outSoundSetup.right.dac_mute = 0;


// CD-ROM maps to aux1, 4bit->5bit, no mute control
	theValue = this->ReadReg(SBPRO_CD);
	outSoundSetup.left.aux1_mix_gain = 0x1E - ((theValue >> 3) & 0x1E);
	outSoundSetup.right.aux1_mix_gain = 0x1E - ((theValue << 1) & 0x1E);
	outSoundSetup.left.aux1_mix_mute = 0;
	outSoundSetup.right.aux1_mix_mute = 0;


// MIDI maps to aux2, 4->5
	theValue = this->ReadReg(SBPRO_MIDI);
	outSoundSetup.left.aux2_mix_gain = 0x1E - ((theValue >> 3) & 0x1E);
	outSoundSetup.right.aux2_mix_gain = 0x1E - ((theValue << 1) & 0x1E);
	outSoundSetup.left.aux2_mix_mute = 0;
	outSoundSetup.right.aux2_mix_mute = 0;


// set line levels, 4->5
	theValue = this->ReadReg(SBPRO_LINE);
	outSoundSetup.left.line_mix_gain = 0x1E - ((theValue >> 3) & 0x1E);
	outSoundSetup.right.line_mix_gain = 0x1E - ((theValue << 1) & 0x1E);
	outSoundSetup.left.line_mix_mute = 0;
	outSoundSetup.right.line_mix_mute = 0;


// only support 22.05kHz
	outSoundSetup.sample_rate = kHz_44_1;
// only support 8 bit stereo playback
	outSoundSetup.playback_format = linear_16bit_little_endian_stereo;
// only support 8 bit stereo capture
	outSoundSetup.capture_format = linear_8bit_unsigned_stereo;
// no control for dither_enable
	outSoundSetup.dither_enable = 0;
// no control for loop_attn
	outSoundSetup.loop_attn = 0x40;
// no control for loop back
	outSoundSetup.loop_enable = 0;
// no control for output boost
	outSoundSetup.output_boost = 0;
// no control for high pass filter
	outSoundSetup.highpass_enable = 0;
// no control for speaker gain
	outSoundSetup.mono_gain = 0x40;
// un/mute speaker
	outSoundSetup.mono_mute = 0;

	return B_OK;
	}

void N505VE_Audio_DevicePCM::HandleInterrupt()
	{
// A DMA transfer to the sound chip just completed, so we update our notions of
// the total samples transferred and the system time at which that happened.
	fSampleClock += fDoubleBufferSize;
	fSampleTime = ::system_time();

// Acknowledge the interrupt by reading the status register.
	this->ReadIO(SBPRO_READ_BUFFER_STATUS);

	if (::atomic_add(&fCount_WriteWaiting, -1) > 0)
		{
// If we have someone waiting to write, release our write semaphore and let them
// proceed. This is how we gate the flow of data into our double buffer.
		::release_sem_etc(fSem_Write, 1, B_DO_NOT_RESCHEDULE);
		}
	else
		{
		::atomic_add(&fCount_WriteWaiting, 1);
// If no one is waiting to write, zero out the buffer we just played. This could
// be made more efficient by maintaining an 'isSilent' flag for each buffer. However,
// this driver is primarily intended for use by the media_server which continuously
// feeds samples to us. Thus this branch is only ever taken when the media_server
// is not operating, but when the driver is still open, which only happens at
// startup and shutdown, and then only for a fraction of a second.
		::memset(fDoubleBuffer[fDoubleBufferCurrent], 0x80, fDoubleBufferSize);
		}
	fDoubleBufferCurrent = 1 - fDoubleBufferCurrent;	
	}

void N505VE_Audio_DevicePCM::IncrementInterruptHandler()
	{
	if (::atomic_add(&fCountInterruptHandler, 1) == 0)
		::install_io_interrupt_handler(fIRQ, sHandleInterrupt, this, 0);
	}

void N505VE_Audio_DevicePCM::DecrementInterruptHandler()
	{
	if (::atomic_add(&fCountInterruptHandler, -1) == 1)
		::remove_io_interrupt_handler(fIRQ, sHandleInterrupt, this);
	}

int32 N505VE_Audio_DevicePCM::sHandleInterrupt(void* inCookie)
	{
	N505VE_Audio_DevicePCM* theDevice = reinterpret_cast<N505VE_Audio_DevicePCM*>(inCookie);
	cpu_status priorStatus = theDevice->DisableInterruptsAndAcquireSpinLock();
	theDevice->HandleInterrupt();
	theDevice->ReleaseSpinLockAndRestoreInterrupts(priorStatus);
	return B_INVOKE_SCHEDULER;
	}

#define DMA_BLOCK 0x80 // block transfer
#define DMA_SINGLE 0x40 // single transfer
#define DMA_DECR 0x20 // address decr
#define DMA_AUTO 0x10	// autoinit
#define DMA_READ 0x04	// mem -> io
#define DMA_WRITE 0x08	// io -> mem

void N505VE_Audio_DevicePCM::StartDMA()
	{
	physical_entry thePhysicalEntry;
	::get_memory_map(fWriteBuffer, fDoubleBufferSize * 2, &thePhysicalEntry, 1);
	fISAModuleInfo->start_isa_dma(fDMA8, thePhysicalEntry.address, thePhysicalEntry.size, DMA_SINGLE | DMA_AUTO | DMA_WRITE, 0);

	cpu_status priorStatus = this->DisableInterruptsAndAcquireSpinLock();

// Enable stereo
	this->WriteReg(SBPRO_OUTPUT_MUX, 0x23);

// Desired sample rate is 44.1kHz. If you work this calculation backwards you'll
// find that in fact the rate is 47.619kHz, which is why we have to maintain separate
// realtime and sample clocks -- tagging the audio buffers delivered to us with those
// values lets the media_server deliver resampled audio at our *actual* playback rate.
	this->WriteData(SBPRO_TIME_CONSTANT);
	this->WriteData(256 - ((1000000 + 44100) / ( 2 * 44100)));

// DMA transfer size -- the count is one less than the amount we actually want moved.
	this->WriteData(SBPRO_TRANSFER_SIZE);
	this->WriteData((fDoubleBufferSize-1)&0xff);
	this->WriteData((fDoubleBufferSize-1)>>8);

// Start playback (trigger the DMA)
	this->WriteData(SBPRO_PLAYBACK_8_BIT);

// Enable the speaker
	this->WriteData(SBPRO_SPEAKER_ENABLE);

	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);
	}

void N505VE_Audio_DevicePCM::StopDMA()
	{
	::memset(fWriteBuffer, 0x80, fDoubleBufferSize*2);

	cpu_status priorStatus = this->DisableInterruptsAndAcquireSpinLock();
	this->WriteData(SBPRO_SPEAKER_DISABLE);
	this->WriteData(SBPRO_EXIT_DMA);
	this->ReleaseSpinLockAndRestoreInterrupts(priorStatus);
	}

status_t N505VE_Audio_DevicePCM::WriteData(uint8 inValue)
	{
	for (int i = 33; i > 0; --i)
		{
		if (this->ReadIO(SBPRO_WRITE_BUFFER_STATUS) & 0x80)
			continue;
		this->WriteIO(SBPRO_WRITE_DATA, inValue);
		return B_OK;
		}
	return B_ERROR;
	}

status_t N505VE_Audio_DevicePCM::ReadData(uint8& outValue)
	{
	for (int i = 33; i > 0; --i)
		{
		if (this->ReadIO(SBPRO_READ_BUFFER_STATUS) & 0x80)
			{
			outValue = this->ReadIO(SBPRO_READ_DATA);
			return B_OK;
			}
		}
	return B_ERROR;
	}

void N505VE_Audio_DevicePCM::WriteReg(uint8 inReg, uint8 inValue)
	{
	this->WriteIO(SBPRO_REG, inReg);
	this->WriteIO(SBPRO_DATA, inValue);
	}

uint8 N505VE_Audio_DevicePCM::ReadReg(uint8 inReg)
	{
	this->WriteIO(SBPRO_REG, inReg);
	return this->ReadIO(SBPRO_DATA);
	}

void N505VE_Audio_DevicePCM::WriteIO(int inOffset, uint8 inValue)
	{
	fISAModuleInfo->write_io_8(fPort + inOffset, inValue);
	}

uint8 N505VE_Audio_DevicePCM::ReadIO(int inOffset)
	{
	return fISAModuleInfo->read_io_8(fPort + inOffset);
	}

cpu_status N505VE_Audio_DevicePCM::DisableInterruptsAndAcquireSpinLock()
	{
	cpu_status priorStatus = ::disable_interrupts();
	::acquire_spinlock(&fSpinLock);
	return priorStatus;
	}

void N505VE_Audio_DevicePCM::ReleaseSpinLockAndRestoreInterrupts(cpu_status inPriorStatus)
	{
	::release_spinlock(&fSpinLock);
	::restore_interrupts(inPriorStatus);
	}
